module wei_cig
  !! compute CIGs
  !!        offset domain (weihcig)
  !! ray parameter domain (weipcig)

#ifndef USE_FFTW
use fftw_sep_fake_mod
#endif


  use wei_util
  use wei_external
  use wei_run
  use wei_down
  use wei_flags
  use wei_tds
  use wei_attr
  !---------------
  implicit none

  real,               private :: pcigeps

contains
  !----------------------------------------------------------------  
!!$=head1 NAME
!!$
!!$weihcig_init -initialize offset imaging operator
!!$
!!$=head1 SYNOPSIS
!!$
!!$ierr= weihcig_init()
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Initializes migration operator whose image (R) domain is offset
!!$
!!$
!!$=head1 LIBRARY
!!$
!!$sepweif90
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<weipcig_init> L<weihcig>
!!$
!!$=cut
!!$

  !! h-domain CIG
  integer function weihcig_init() result(st)
#ifdef DEBUG
    character(len=128), parameter :: unit='weihcig_init'
    if(run%debug) call in(unit)
#endif
    !! do nothing
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weihcig_init


!!$=head1 NAME
!!$
!!$weihcig - Perform offset domain imaging 
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr=weihcig(wfld,iws,ith)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item iws - integer
!!$
!!$  current frequency slice
!!$
!!$=item ith - integer
!!$
!!$  current thread
!!$
!!$=item wfld - complex(:,:,:,:,:)
!!$
!!$  wavefield
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Imaging operation when in the offset domain
!!$
!!$
!!$=head1 LIBRARY
!!$
!!$sepweif90
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<weihcig_init> L<weipcig>
!!$
!!$=cut




  !----------------------------------------------------------------
  integer function weihcig(wfld,iws,ith) result(st)
    complex, dimension(:,:,:,:,:), pointer :: wfld
    integer, intent(in) ::iws,ith
#ifdef DEBUG
    character(len=128), parameter :: unit='weihcig'
    if(run%depar) call in(unit)
#endif
    !! do nothing
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weihcig
  !----------------------------------------------------------------  
!!$=head1 NAME
!!$
!!$weipcig_init -initialize offset ray parameter imaging
!!$
!!$=head1 SYNOPSIS
!!$
!!$ierr= weihcig_init()
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Initializes migration operator whose image (R) domain is p_h
!!$
!!$ Grabs pcig_eps (default=1.0)
!!$
!!$
!!$=head1 LIBRARY
!!$
!!$sepweif90
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<weihcig_init> L<weipcig>
!!$
!!$=cut
!!$

  !----------------------------------------------------------------  
  !! p-domain CIG
  integer function weipcig_init() result(st)
    integer, dimension(3) :: ni
    logical               :: fftmeasure
    integer               :: fftsetup
    integer               :: er,i
    integer               :: ipx,ipy, jpx,jpy, ihx,ihy
    real                  :: arg
#ifdef DEBUG
    character(len=128), parameter :: unit='weipcig_init'
    if(run%debug) call in(unit)
#endif



#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weipcig_init
!!$=head1 NAME
!!$
!!$weipcig - Perform offset ray parameter imaging 
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr=weipcig(wfld,iws,ith)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item iws - integer
!!$
!!$  current frequency slice
!!$
!!$=item ith - integer
!!$
!!$  current thread
!!$
!!$=item wfld - complex(:,:,:,:,:)
!!$
!!$  wavefield
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Imaging operation when in the offset ray parameter (p_h) domain
!!$
!!$
!!$=head1 LIBRARY
!!$
!!$sepweif90
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<weipcig_init> L<weihcig>
!!$
!!$=cut
  !----------------------------------------------------------------
  integer function weipcig(wfld,iws,ith) result(st)
    complex, dimension(:,:,:,:,:), pointer :: wfld
    complex, dimension(:,:,:,:) :: midin
    complex, dimension(:,:) :: exp_multx,exp_multy
    integer, intent(in) ::iws,ith
		real :: w,px,py,hx,hy
		complex :: shift
		integer :: ipx,ipy,imx,imy,ihy,ihx
#ifdef DEBUG
    character(len=128), parameter :: unit='weipcig'
    if(run%depar) call in(unit)
#endif

		allocate(midin(down%wsep%n(1),down%wsep%n(2),down%wsep%n(3),down%wsep%n(4)),exp_multx(down%wsep%n(3),down%rsep%n(3)),exp_multy(down%wsep%n(4),down%rsep%n(4)))
		midin=0.0

    w=down%w_local(iws)

		do ipx=1,down%rsep%n(3)
			px=(ipx-1)*down%rsep%d(3)+down%rsep%o(3)
			do ihx=1,down%wsep%n(3)
				hx=(ihx-1)*down%wsep%d(3)+down%wsep%o(3)
				exp_multx(ihx,ipx)=exp(cmplx(0,-w*px*hx))
			end do
		end do


    if(down%adj) then !! MIGRATION
		
			midin=wfld(:,:,:,:,ith)
			wfld(:,:,:,:,ith)=0.0
			!! get the p_hx axis done
					do ipx=1,down%rsep%n(3)
						do ihy=1,down%wsep%n(4)
							do ihx=1,down%wsep%n(3)
			do imy=1,down%wsep%n(2)
				do imx=1,down%wsep%n(1)
								wfld(imx,imy,ipx,ihy,ith)=wfld(imx,imy,ipx,ihy,ith)+midin(imx,imy,ihx,ihy)*exp_multx(ihx,ipx)
							end do
						end do
					end do
				end do
			end do

!	if(run%m_ith==0) write(0,*)"sum wfld2",sum(wfld)

!if(1==3) then
!	if(down%wsep%n(4) > 1) then
!			do ipy=1,down%rsep%n(4)
!				py=(ipy-1)*down%rsep%d(4)+down%rsep%o(4)
!				do ihy=1,down%wsep%n(4)
!					hy=(ihy-1)*down%wsep%d(4)+down%wsep%o(4)
!					exp_multy(ihy,ipy)=exp(cmplx(0,-w*py*hy))
!				end do
!			end do
!
!			!! get the p_hy axis done -- NEVER TESTED
!			do imy=1,down%wsep%n(2)
!				do imx=1,down%wsep%n(1)
!					midin(:,:)=wfld(imx,imy,:,:,ith)
!					wfld(imx,imy,:,:,ith)=0.0
!					do ipy=1,down%rsep%n(4)
!						do ihx=1,down%wsep%n(3)
!							do ihy=1,down%wsep%n(4)
!								wfld(imx,imy,ihx,ipy,ith)=wfld(imx,imy,ihx,ipy,ith)+midin(ihy,ipy)*exp_multy(ihy,ipy)
!							end do
!						end do
!					end do
!				end do
!			end do
!	end if
!end if
!	if(run%m_ith==0) write(0,*)"sum wfld3",sum(wfld)

    else             !! MODELING
					midin=wfld(:,:,:,:,ith)
					wfld(:,:,:,:,ith)=0.0

			!! get the p_hx axis done
					do ipx=1,down%rsep%n(3)
						do ihy=1,down%wsep%n(4)
							do ihx=1,down%wsep%n(3)
			do imy=1,down%wsep%n(2)
				do imx=1,down%wsep%n(1)
								wfld(imx,imy,ihx,ihy,ith)=wfld(imx,imy,ihx,ihy,ith)+midin(imx,imy,ipx,ihy)*conjg(exp_multx(ihx,ipx))
							end do
						end do
					end do
				end do
			end do

!if(1==3) then
!	if(down%wsep%n(4) > 1) then
!			do ipy=1,down%rsep%n(4)
!				py=(ipy-1)*down%rsep%d(4)+down%rsep%o(4)
!				do ihy=1,down%wsep%n(4)
!					hy=(ihy-1)*down%wsep%d(4)+down%wsep%o(4)
!					exp_multy(ihy,ipy)=exp(cmplx(0,w*py*hy))
!				end do
!			end do
!
!			!! get the p_hy axis done -- NEVER TESTED
!			do imy=1,down%wsep%n(2)
!				do imx=1,down%wsep%n(1)
!					midin(:,:)=wfld(imx,imy,:,:,ith)
!					wfld(imx,imy,:,:,ith)=0.0
!					do ipy=1,down%rsep%n(4)
!						do ihx=1,down%wsep%n(3)
!							do ihy=1,down%wsep%n(4)
!								wfld(imx,imy,ihx,ipy,ith)=wfld(imx,imy,ihx,py,ith)+midin(ihx,ipy)*exp_multy(ihy,ipy)
!							end do
!						end do
!					end do
!				end do
!			end do
!		end if
!end if

    end if
 
!  if(run%m_ith==0) write(0,*) "CHECK OUT",sum(cabs(wfld)),sum(exp_multx)
		deallocate(midin,exp_multx,exp_multy)

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weipcig
  !----------------------------------------------------------------  

  !----------------------------------------------------------------
end module wei_cig

